'use strict';

const fs = require('fs');
const path = require('path');
const Sequelize = require('sequelize');
const mysql = require('mysql');
const toUnnamed = require('named-placeholders')();
const originalQuery = require('mysql/lib/Connection').prototype.query;

const uuidv4 = require('uuid/v4');

require('mysql/lib/Connection').prototype.query = function (...args) {
  let query = args[0];

  if (!query.values || Array.isArray(query.values)) {
    return originalQuery.apply(this, args);
  }
  
  const paramsMap = toUnnamed(query.sql, query.values); 
  query.sql = paramsMap[0];
  query.values = paramsMap[1];
  
  return originalQuery.apply(this, args);
};


//Singleton implementation
const FINECT_DB_KEY = Symbol.for("finect.api.db");

// check if the global object has this symbol
// add it if it does not have the symbol, yet
// ------------------------------------------
let globalSymbols = Object.getOwnPropertySymbols(global);
if (!(globalSymbols.indexOf(FINECT_DB_KEY) > -1)) {

  // pre-checks
  const valid = process.env.MYSQL_DB && process.env.MYSQL_USER && process.env.MYSQL_PASS && process.env.MYSQL_HOST && process.env.MYSQL_PORT;
  if (!valid) throw new Error('Missing required environment variables');

  const sequelize = new Sequelize(process.env.MYSQL_DB, process.env.MYSQL_USER, process.env.MYSQL_PASS, {
    host: process.env.MYSQL_HOST,
    dialect: 'mysql',
    operatorsAliases: false,

    pool: {
      max: 10,
      min: 0,
      acquire: 30000,
      idle: 10000
    }
    // logging: logger.info
  });

  let db = {};

  const models = global.__modelsdir || path.join(global.__basedir, 'api/controllers');
  walk(models)
    .forEach(file => {
      const model = sequelize['import'](file);
      db[model.name] = model;
    });
  
  Object.keys(db).forEach(modelName => {
    if (db[modelName].associate) {
      db[modelName].associate(db);
    }
  });

  db.executeQuery = (model, definition, callback) => {
    const reader = getReader();
    const parameters = model.parameters ? model.parameters(definition.params) : undefined;

    reader.query(model.sql, parameters,
      (error, results) => {
        if (error) 
        {
          callback(error);
          return;
        }
        
        // CALL SP
        // if (Array.isArray(results) && results.length == 2 && results[1].affectedRows !== undefined) {
        //   results = results[0];
        // }

        callback(undefined, model.map ? model.map(results, definition) : results);
      });
  }

  db.executeQueryAsync = (model, definition) => {
    return new Promise(resolve => {
      const reader = getReader();

      const parameters = model.parameters ? model.parameters(definition.params) : undefined;
      const sql = typeof model.sql === 'function' ? model.sql(definition.params) : model.sql;
      
      reader.query(sql, parameters,
        (error, results) => {
          if (error) 
          {
            resolve({ error });
            return;
          }
          
          resolve({ data: model.map ? model.map(results, definition) : results });
        });  
    });
  }

  db.uniqid = () => {
    return uuidv4().replace(/-/g, "");
  };

  db.sequelize = sequelize;
  db.Sequelize = Sequelize;

  global[FINECT_DB_KEY] = {
    db: db 
  }
}

let poolReader;
function getReader() {
  //logger.debug(config.db.reader.host);
  if (poolReader) return poolReader;

  poolReader = mysql.createPool({
    multipleStatements : true,
    connectionLimit : 10,
    aquireTimeout: process.env.MYSQL_TIMEOUT || 30000,
    host: process.env.MYSQL_HOST,
    port: process.env.MYSQL_PORT,
    user: process.env.MYSQL_USER,
    password : process.env.MYSQL_PASS,
    database: process.env.MYSQL_DB,
    timezone: 'utc',
    typeCast: function castField( field, useDefaultTypeCasting ) {

      // We only want to cast bit fields that have a single-bit in them. If the field
      // has more than one bit, then we cannot assume it is supposed to be a Boolean.
      if (field.length > 1) return( useDefaultTypeCasting() );

      const bytes = field.buffer();
      switch(field.type) {
        case 'LONG':
        case 'TINY':
          return( String.fromCharCode(bytes[0]) !== '0' );

        case 'BIT':
          return( bytes[0] === 1 );

        default: 
          return( useDefaultTypeCasting() );
      }
    }
  });

  return poolReader;
}

function walk(dir) {
  var results = []
  var list = fs.readdirSync(dir)
  list.forEach(function(file) {

    file = dir + '/' + file;
    let stat = fs.statSync(file);
    if (stat && stat.isDirectory()) results = results.concat(walk(file));
    else {
      if (file.indexOf('model.js') == -1) return;
      results.push(file);
    }
  });

  return results
}

let singleton = {};
Object.defineProperty(singleton, "instance", {
  get: function(){
    return global[FINECT_DB_KEY];
  }
});

// ensure the API is never changed
// -------------------------------
Object.freeze(singleton);

// export the singleton API only
// -----------------------------
module.exports = singleton.instance.db;
