"use strict";

const logger = require('@finect/logger');
const jsf = require('json-schema-faker');
const queryString = require('query-string');
const jwt = require('jsonwebtoken');

jsf.option({
    alwaysFakeOptionals: true
});

jsf.extend('faker', () => {
    const faker = require('faker');

    faker.locale = "en";
    // faker.custom = {
    //   number: (length, a, b, c) => {
    //     return length;
    //   }
    // };
    return faker;
});

module.exports.errors = (err, req, res, next) => {
    if (err.failedValidation) err.statusCode = 400;
    res.toError(err.statusCode || 500, "-1", err);

    next();
};

module.exports.ignoreUnderscoreJsonProperties = (key, value) => {
    return key.startsWith('__') ? undefined : value;
};

module.exports.response = (req, res, next) => {

    res.toError = (code, internalCode, err) => {
        if (code === 403) {
            code = 401; 
            err = "Unauthorized";
        } 
        else {
            err = err.stack || err.message || err || 'Internal server error.'
            logger.error(err);
        }

        let result = {
            code: code,
            internalCode: internalCode
        };

        if (process.env.NODE_ENV !== 'production') {
            result.message = err;
        }

        res.status(code).send(result).end();
    },

    res.isErrorOrEmptyResult = (err, data, message) => {
        
        if (err) {
            res.toError(500, '-100', 'SQL-ERROR: ' + err.message);
            return true;
        }

        if (data == undefined) {
          res.toError(404, '-2', message);
          return true;
        }  

        return false;
    },    

    res.notImplemented = () => {

        let schema = req.swagger.path[req.method.toLowerCase()].responses['501'].schema;
        if (!schema) {
            res.status(501).json().end();
            return;
        } 

        jsf.resolve(schema).then(function(result) {
            result.code = 501;
            res.status(501).json(result).end(); 
        });
    },

    res.toResult = (data) => {
        res.json({
            code: 200,
            data: data
        });
    },   

    res.toPagedResult = (data, total) => {  
        const offset = req.swagger.params.offset.value;
        let result = {
            code: 200,
            paging: {
                totalItems: total,
                limit: req.swagger.params.limit.value
            },
            data: data || []
        };
           
        result.paging.links = [];
        req.query.offset = offset;
        req.query.limit = req.swagger.params.limit.value;

        if (offset + req.query.limit < total) {

            req.query.offset = offset + req.query.limit + 1;
            result.paging.links.push({
                rel: 'next',
                href: req.path + '?' + queryString.stringify(req.query)
            });
        }
      
        if (offset > 0) {

            req.query.offset = offset;
            req.query.limit *= -1;
            result.paging.links.push({
                rel: 'prev',
                href: req.path + '?' + queryString.stringify(req.query)
            });
        }

        res.json(result);
    };

    res.toCursorPagedResult = (data, total, sinceId, maxId) => {  
        let result = {
            code: 200,
            paging: {
                total: total
            },
            data: data
        };
            
        result.paging.links = [];
        req.query.limit = req.swagger.params.limit.value;

        req.query.offset = maxId;
        result.paging.links.push({
            rel: 'next',
            href: req.path + '?' + queryString.stringify(req.query)
        });

        req.query.offset = sinceId;
        req.query.limit *= -1;
        result.paging.links.push({
            rel: 'prev',
            href: req.path + '?' + queryString.stringify(req.query)
        });

        res.json(result);
    };

    
	req.getCookies = () => {
		let cookies = {};
		req.headers && req.headers.cookie && req.headers.cookie.split(';').forEach(function(cookie) {
			const parts = cookie.match(/(.*?)=(.*)$/);
			cookies[ parts[1].trim() ] = (parts[2] || '').trim();
		});
	
		return cookies;
	}

	req.getCookie = (name) => {
		let cookies = req.getCookies();
		return cookies[name];
	}

	req.isAuthorized = () => {		
        let token;
		if (req.headers.authorization) {
			token = req.headers.authorization.split(' ')[1];
		}

		if (!token) {
			token = req.getCookie(process.env.FINECT_COOKIE || '_fi_jwt');
		}

		if (!token) return false;

		const decoded = jwt.decode(token);
		if (decoded) {
            req.swagger.params['x-authenticated-userid'] = decoded.sub;
            return true;    
        }

        if (req.headers.authorization && req.headers['x-authenticated-userid']) {
            req.swagger.params['x-authenticated-userid'] = req.headers['x-authenticated-userid'];             
            
            return true;
        }

        return false;
	}

    next();
};