'use strict';

const _ = require('lodash');

let defaultIdConventions = ['id'];
let globalCache;

let _schema;

const relationType = {
    collection: 0,
    association: 1
};

const associationType = {
    type: relationType.association,
    separator: '_'
};

const collectionType = {
    type: relationType.collection,
    separator: '__'
};


module.exports.setIdentifierConventions = idConventions => {
    defaultIdConventions = idConventions;
}

module.exports.mapDynamic = (arr, schema, property) =>
{
    if (arr.length == 0) return [];

    _schema = undefined;
    if (schema) {
        schema = resolveAllOf(schema);
        _schema = property ? findPropertyInSchema(schema, property) : schema;
    }

    globalCache = {};
    
    arr.map(row => {
        injectResultInCollection(row, 'root');
    });

    return globalCache['root'];
} 

module.exports.mapDynamicToOne = (arr, schema, property) =>
{
    const result = module.exports.mapDynamic(arr, schema, property);
    return result.length == 0 ? undefined : result[0];
} 

function injectResultInCollection(row, key)
{
    globalCache[key] = globalCache[key] || [];
    let mappedObject = _.find(globalCache[key], cache => { 
        return _.find(defaultIdConventions, id => {
            return row[id] !== undefined && cache[id] === row[id] || 
                (cache[id] instanceof Date && row[id] instanceof Date && cache[id].getTime()  === row[id].getTime());
        });
    });  

    const inCache = mappedObject !== undefined;
    if (!mappedObject)
    {
        mappedObject = _schema ? getPropertiesBySchema(row, findPropertyInSchema(_schema, key)) : getProperties(row);
        globalCache[key].push(mappedObject);
    }

    injectResultInObject(row, mappedObject);
    return mappedObject;
}

function injectResultInObject(row, mappedObject) 
{
    // copy collection
    let collections = _.transform(row, (result, value, key) => {
        const idxAssoc = key.indexOf(associationType.separator);
        const idxColl =  key.indexOf(collectionType.separator);

        if (idxColl == -1) return;
        if (idxAssoc >= 0 && idxAssoc < idxColl) return;
        result[key] = value;
    }, {});

    createMappedObject(collections, collectionType).forEach(result => {
        if (!mappedObject[result.field]) mappedObject[result.field] = [];

        if (isEmpty(result.obj)) return;
        mappedObject[result.field].push(result.obj);        
    });

    // Copy associations
    let associations = _.transform(row, (result, value, key) => {
        const idxAssoc = key.indexOf(associationType.separator);
        const idxColl =  key.indexOf(collectionType.separator);

        if (idxAssoc == -1) return;
        if (idxColl >= 0 && idxColl <= idxAssoc) return;
        result[key] = value;
    }, {});

    createMappedObject(associations, associationType).forEach(result => {
        if (isEmpty(result.obj)) return;
        mappedObject[result.field] = result.obj;        
    }); 
}

function createMappedObject(collections, relation)
{
    let result = [];

    if (Object.keys(collections).length > 0) {
        const colNames = _.transform(collections, (result, value, key) => {
            const colName = key.substring(0, key.indexOf(relation.separator));

            if (result.indexOf(colName) === -1) result.push(colName);
        }, []);
        
        colNames.map(col => {

            let innerCollection = _.transform(collections, (result, value, key) => {
                if (!key.startsWith(col+relation.separator)) return;
                result[key.replace(col+relation.separator, '')] = value;
            });

            const collection = injectResultInCollection(innerCollection, col); 
            result.push({ field: col, obj: collection });

        });
    }

    return result;
}   

function getPropertiesBySchema(row, objSchema)
{
    if (!objSchema) objSchema = _schema;

    return _.transform(objSchema.properties, (result, value, key) => {
        result[key] = row[key] === undefined ? null : row[key];
    }, {});
}

function getProperties(row)
{
    return _.transform(row, (result, value, key) => {
        if (key.includes('_')) return;
        result[key] = value;
    }, {});
}

function resolveAllOf(inputSpec){
    let out; 
    
    if (inputSpec && typeof inputSpec === 'object') {
        if (inputSpec.allOf) {
            let allOf = inputSpec.allOf;
            delete inputSpec.allOf;
            let nested = _.mergeWith({}, ...allOf, customizer);
            out = _.defaultsDeep(inputSpec, nested, customizer);

            if (out.allOf) out = resolveAllOf(out)
        } 
        else out = inputSpec;

        Object.keys(out).forEach((key, context) => {
            out[key] = resolveAllOf(out[key]);
        });
    } 
    else out = inputSpec;

    return inputSpec;
};

function customizer(objValue, srcValue) {
    if (_.isArray(objValue)) {
        return _.union(objValue, srcValue);
    }
};

function isEmpty(object)
{
    return Object.values(object).every(x => (x === null || x === ''));
}

function findPropertyInSchema(object, key) {
    var value;
    Object.keys(object).some(k => {
        
        if (k === key) {
            if (object[k].type === 'array' || (Array.isArray(object[k].type) && object[k].type.indexOf('array') > -1))
            {
                value = object[k].items;
                return true;
            }

            if (!object[k].properties) return false
            value = object[k];

            return true;
        }

        if (object[k] && typeof object[k] === 'object') {
            value = findPropertyInSchema(object[k], key);
            return value !== undefined;
        }
    });

    return value;
}
