'use strict';

const Item = require('postman-collection').Item;
const QueryParam = require('postman-collection').QueryParam;
const Events = require('./events');


const vkbeautify = require('vkbeautify');
const _ = require('underscore');

class Endpoint {
    constructor(method, url, def, swaggerCommons)
    {
        this.method = method;
        this.url = url;
        this.def = def;
        this.swaggerCommons = swaggerCommons;

        this.postmanItems = [];
    }

    export(globalTests, security){
        this.globalTests = {
            parameters: globalTests && globalTests['params'] || [], 
            tests: globalTests && globalTests['tests'] || [] 
        };

        this.security = security;

        buildPostmanItems(this);
        return this.postmanItems; 
    }
}

function buildPostmanItems(endpoint)
{
    const produces = endpoint.def.produces || endpoint.swaggerCommons.produces;
    const consumes = endpoint.def.consumes || endpoint.swaggerCommons.consumes;

    consumes.forEach(content => {
        produces.forEach(accept => {
            _.each(endpoint.def.responses, (response, status) =>
            {
                let tests = response['x-pm-test'] || [];
                if (tests.length == 0 && status !== 'default') tests = buildDefaultsTest(endpoint, status);
    
                let globalParameters = endpoint.globalTests.parameters.filter((p) => {
                    let responses = p.except && p.except.responses || [];
                    let methods = p.except && p.methods || [];
    
                    return !responses.map(String).includes(status) && !methods.includes(endpoint.method); 
                });
    
                let globalTests = getGlobalsTest(endpoint, status);             
                tests.forEach(test => {
                    test.params = (test.params || []).concat(globalParameters);
                    test.raw = (test.raw || '') + globalTests;

                    if (status !== '401') test.params.push(endpoint.security.param);
                    endpoint.postmanItems.push(buildPostmanItem(endpoint.method, endpoint.url, test, content, accept, status));
                });
            });
        });
    });
}

function getGlobalsTest(endpoint, status)
{
    let raw = '';
    endpoint.globalTests.tests.forEach(test => {
        if (!test.except.responses.map(String).includes(status) && !test.except.methods.includes(endpoint.method))
        {
            raw += test.raw;
        }
    });

    return raw;
}

function buildDefaultsTest(endpoint, status)
{
    let tests = [
        {
            "description": 'I can get ['+ status + ']' + ' on ' + endpoint.url,
        }  
    ];

    if (!endpoint.def.parameters) return tests;

    tests[0].params = [];
    endpoint.def.parameters.forEach(parameter => {
        if (parameter.required)
            tests[0].params.push({
                name: parameter.name,
                in: parameter.in,
                value: parameter['x-pm-test-value'] || parameter.default || parameter.minimum || defaultVal(parameter)
            });
    });

    return tests;
}

function buildPostmanItem(method, url, test, content, accept, status)
{
    let queryString = _.where(test.params, { in: 'query' })
        .map(p => p.name + '=' +p.value).join('&');

    let headerParam = _.where(test.params, { in: 'header'});
    let pathParameters = _.where(test.params, {in: 'path'});
    let urlVariables = [];

    let bodyParam = _.find(test.params, { in: 'body'});

    pathParameters.forEach(p => {
        urlVariables.push({ key: p.name, value: p.value });
        url = url.replace('{'+p.name+'}', ':'+p.name);
    });

    let item = new Item({
        name: (test.description || 'I can get ['+ status + ']' + ' on ' + url).replace('[url]', url),
        request: {
            method: method,
            url: '{{base-url}}'+ url + (queryString ? '?'+ queryString : ''),
            header: [
                { key: 'accept', value: accept },
                { key: 'content-type', value: content }
            ]
        }
    });

    if (headerParam.length > 0) {
       headerParam.forEach(header => {
            item.request.headers.members.push({key: header.name, value: header.value});
       });  
    }

    if (urlVariables.length > 0) item.request.url.variable = urlVariables;

    if (bodyParam) {
        item.request.body = { mode: 'raw', raw: undefined };

        if (content.endsWith('xml')) item.request.body.raw = vkbeautify.xml(bodyParam[content]);
        else if (content.endsWith('json')) item.request.body.raw = vkbeautify.json(bodyParam[content]);
        else item.request.body.raw = bodyParam[content];
    }   

    item.events.add(Events.getTests(accept, status, test.raw));

    return item;
}

function defaultVal(parameter) {
    if (parameter.in === 'body') return '';

    let type = parameter.items ? parameter.items.type : parameter.type;    
    if (typeof type !== 'string') throw new TypeError('Type must be a string.');

    // Handle simple types (primitives and plain function/object)
    switch (type) {
        case 'boolean'   : return 'false';
        case 'integer'   : 
        case 'number'    : return '0';
        case 'string'    : {
            // TODO: check format (date, date-time, enums, etc.)
            return 'string'; 
        }
        default : return '';
    }
}

module.exports = Endpoint;